/*-------------------------------------------------------------------
 *
 *  @file names.c
 *
 *  @author Collin R. Mulliner <collin@mulliner.org>
 *
 *  (c) Collin R. Mulliner
 *
 *  This is part of the Bluetooth Location Tracker (BLT) project.
 *
 *  --
 *
 *  This is the central DEVICE NAME storage, the storage is 
 *  implemented twice, as a fixed size hash table and as a dynamic 
 *  binary search tree (selected one of both in names.h)
 *
 *-------------------------------------------------------------------*/

/*
 * This file is part of blt_server
 *
 * blt_server is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * blt_server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blt_server; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "blt_server.h"

#ifdef DN_TREE
int blt_dn_walkTree(struct tnode_t *t, FILE *fp, int WorF)
{
	char buffer[1024];
	

	if (t->r != NULL) blt_dn_walkTree(t->r, fp, WorF);
	if (t->l != NULL) blt_dn_walkTree(t->l, fp, WorF);

	if (WorF) {
		printWithEsc(t->dev.name, buffer, 1024);
		fprintf(fp, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X \"%s\"\n",
			t->dev.bdAddr[0],t->dev.bdAddr[1],t->dev.bdAddr[2],
			t->dev.bdAddr[3],t->dev.bdAddr[4],t->dev.bdAddr[5],
			buffer);
	}
	else {
		if (t->r != NULL) free(t->r);
		if (t->l != NULL) free(t->l);
	}
	return(1);
}

void blt_dn_init(struct blt_dev_names_t *dn)
{
	dn->numDevices = 0;
	dn->devices = NULL;
	pthread_rwlock_init(&dn->lock, NULL);
}

void blt_dn_destroy(struct blt_dev_names_t *dn)
{
	if (dn->devices != NULL) {
		blt_dn_walkTree(dn->devices, NULL, 0);
		free(dn->devices);
		dn->devices = NULL;
	}
	pthread_rwlock_destroy(&dn->lock);
}

int blt_dn_write(struct blt_dev_names_t *dn, struct blt_device_t *nDev)
{
	struct tnode_t *tree = dn->devices;
	int pos = 0;
	

	/* write lock */
	pthread_rwlock_wrlock(&dn->lock);
	
	/* empty tree */	
	if (dn->devices == NULL) {
		dn->devices = malloc(sizeof(struct tnode_t));
		dn->devices->l = NULL;
		dn->devices->r = NULL;
		memcpy(&dn->devices->dev, nDev, sizeof(struct blt_device_t));
		dn->numDevices++;
		
		/* unlock */
		pthread_rwlock_unlock(&dn->lock);
		
		#ifdef DEBUG3
		fprintf(stderr, "new device: %.2X:%.2X:%.2X:%.2X:%.2X:%.2X name: %s\n",
			nDev->bdAddr[0],nDev->bdAddr[1],nDev->bdAddr[2],nDev->bdAddr[3],
			nDev->bdAddr[4],nDev->bdAddr[5], nDev->name);
		#endif
		
		return(1);
	}

	do {
		pos = memcmp(tree->dev.bdAddr, nDev->bdAddr, 6);
		
		/* update */
		if (pos == 0) {		
			#ifdef DEBUG3
			fprintf(stderr, "update device: %.2X:%.2X:%.2X:%.2X:%.2X:%.2X name: %s\n",
				nDev->bdAddr[0],nDev->bdAddr[1],nDev->bdAddr[2],nDev->bdAddr[3],
				nDev->bdAddr[4],nDev->bdAddr[5], nDev->name);
			#endif
			/* memcpy(&tree->dev, nDev, sizeof(struct blt_device_t)); */
			if (nDev->name[0] != 0) memcpy(tree->dev.name, nDev->name, BLT_DEVICE_NAME_LENGTH+1);
			if (nDev->devClass != 0) tree->dev.devClass = nDev->devClass;
			break;
		}
		/* left */
		else if (pos < 0) {
			if (tree->l != NULL) {
				tree = tree->l;
				continue;
			}
			else {
				tree->l = malloc(sizeof(struct tnode_t));
				tree = tree->l;
			}
		}
		/* right */
		else if (pos > 0) {
			if (tree->r != NULL) {
				tree = tree->r;
				continue;
			}
			else {
				tree->r = malloc(sizeof(struct tnode_t));
				tree = tree->r;
			}
		}

		/* insert */
		#ifdef DEBUG3
		fprintf(stderr, "new device: %.2X:%.2X:%.2X:%.2X:%.2X:%.2X name: %s\n",
			nDev->bdAddr[0],nDev->bdAddr[1],nDev->bdAddr[2],nDev->bdAddr[3],
			nDev->bdAddr[4],nDev->bdAddr[5], nDev->name);
		#endif
		dn->numDevices++;
		tree->l = NULL;
		tree->r = NULL;
		memcpy(&tree->dev, nDev, sizeof(struct blt_device_t));
		break;
		
	} while(1);
	
	/* unlock */
	pthread_rwlock_unlock(&dn->lock);
	
	return(1);
}

int blt_dn_read(struct blt_dev_names_t *dn, struct blt_device_t *nDev)
{
	struct tnode_t *tree = dn->devices;
	int pos = 0;
	

	/* read lock */
	pthread_rwlock_rdlock(&dn->lock);
	
	/* empty tree */	
	if (dn->devices == NULL) {
		pthread_rwlock_unlock(&dn->lock);
		return(0);
	}

	do {
		pos = memcmp(tree->dev.bdAddr, nDev->bdAddr, 6);
		
		/* found */
		if (pos == 0) {
			/* memcpy(nDev, &tree->dev, sizeof(struct blt_device_t)); */
			if (nDev->name[0] == 0)	memcpy(nDev->name, tree->dev.name, BLT_DEVICE_NAME_LENGTH+1);
			if (nDev->devClass == 0) nDev->devClass = tree->dev.devClass;
			break;
		}
		/* left */
		else if (pos < 0) {
			if (tree->l != NULL) {
				tree = tree->l;
				continue;
			}
			else {
				pthread_rwlock_unlock(&dn->lock);
				return(0);
			}
		}
		/* right */
		else if (pos > 0) {
			if (tree->r != NULL) {
				tree = tree->r;
				continue;
			}
			else {
				pthread_rwlock_unlock(&dn->lock);
				return(0);
			}
		}		
	} while(1);

	pthread_rwlock_unlock(&dn->lock);	
	return(1);
}

int blt_dn_save(struct blt_dev_names_t *dn)
{
	FILE *fp;	
	
	if ((fp = fopen(BLT_DEVICE_NAMES_FILE, "w")) == NULL) {
		return(0);
	}
	
	if (dn->devices != NULL) {
		blt_dn_walkTree(dn->devices, fp, 1);
	}
	else return(0);
	
	fclose(fp);
	return(1);
}
#endif

#ifdef DN_HASH
void blt_dn_init(struct blt_dev_names_t *dn)
{
	dn->maxNumDevices = DN_HASHTABLE_SIZE;
	dn->numDevices = 0;
	dn->devices = malloc(dn->maxNumDevices * sizeof(struct blt_device_t));
	memset(dn->devices, 0, dn->maxNumDevices * sizeof(struct blt_device_t));
	pthread_rwlock_init(&dn->lock, NULL);
}

void blt_dn_destroy(struct blt_dev_names_t *dn)
{
	if (dn->devices != NULL) free(dn->devices);
	pthread_rwlock_destroy(&dn->lock);
}

int hash(unsigned char *mac, int tableSize)
{
	unsigned int *value;
	value = (unsigned int*) (mac+2);	
	return((*value) % tableSize);
}

int blt_dn_write(struct blt_dev_names_t *dn, struct blt_device_t *nDev)
{
	const unsigned char EMPTY_MAC[6] = { 0x00,0x00,0x00,0x00,0x00,0x00 };
	int pos;
	int bong;
	
	
	/* write lock */
	pthread_rwlock_wrlock(&dn->lock);
	
	pos = hash(nDev->bdAddr, dn->maxNumDevices);
	do {
		/* update */
		if (memcmp(dn->devices[pos].bdAddr, nDev->bdAddr, 6) == 0) {
			#ifdef DEBUG3
			fprintf(stderr, "updated device: %.2X:%.2X:%.2X:%.2X:%.2X:%.2X name: %s\n",
			nDev->bdAddr[0],nDev->bdAddr[1],nDev->bdAddr[2],nDev->bdAddr[3],nDev->bdAddr[4],nDev->bdAddr[5],
			nDev->name);
			#endif
			
			/* memcpy(&dn->devices[pos], nDev, sizeof(struct blt_device_t)); */
			if (nDev->name[0] != 0) memcpy(dn->devices[pos].name, nDev->name, BLT_DEVICE_NAME_LENGTH+1);
			if (nDev->devClass != 0) dn->devices[pos].devClass = nDev->devClass;
			break;
		}
		/* insert */
		else if (memcmp(dn->devices[pos].bdAddr, EMPTY_MAC, 6) == 0) {
			#ifdef DEBUG3
			fprintf(stderr, "new device: %.2X:%.2X:%.2X:%.2X:%.2X:%.2X name: %s\n",
			nDev->bdAddr[0],nDev->bdAddr[1],nDev->bdAddr[2],nDev->bdAddr[3],nDev->bdAddr[4],nDev->bdAddr[5],
			nDev->name);
			#endif
		
			dn->numDevices++;
			memcpy(&dn->devices[pos], nDev, sizeof(struct blt_device_t));
			break;
		}
		/* collision */
		else {
			pos++;
			/* if table end is reached wrap to begin of table */
			if (pos == dn->maxNumDevices) {
				pos = 0;
				if (bong == 0) bong = 1;
				else {
					fprintf(stderr, "ERROR: hash table full!\n");
					pthread_rwlock_unlock(&dn->lock);
					return(0);
				}
			}
		}
	} while (1);
		
	pthread_rwlock_unlock(&dn->lock);
	return(1);
}

int blt_dn_read(struct blt_dev_names_t *dn, struct blt_device_t *nDev)
{
	const unsigned char EMPTY_MAC[6] = { 0x00,0x00,0x00,0x00,0x00,0x00 };
	int pos;
	int bong = 0;


	pthread_rwlock_rdlock(&dn->lock);

	pos = hash(nDev->bdAddr, dn->maxNumDevices);
	do {
		if (memcmp(dn->devices[pos].bdAddr, nDev->bdAddr, 6) == 0) {
			if (nDev->name[0] == 0)	memcpy(nDev->name, dn->devices[pos].name, BLT_DEVICE_NAME_LENGTH+1);
			if (nDev->devClass == 0) nDev->devClass = dn->devices[pos].devClass;
			break;
		}
		else if (memcmp(dn->devices[pos].bdAddr, EMPTY_MAC, 6) == 0) {
			pthread_rwlock_unlock(&dn->lock);
			return(0);
		}
		else {
			pos++;
			/* if table end is reached wrap to begin of table */
			if (pos == dn->maxNumDevices) {
				pos = 0;
				if (bong == 0) bong = 1;
				else {
					fprintf(stderr, "ERROR: hash table full!\n");
					pthread_rwlock_unlock(&dn->lock);
					return(0);
				}
			}
		}
	} while(1);
	
	pthread_rwlock_unlock(&dn->lock);
	return(1);
}

int blt_dn_save(struct blt_dev_names_t *dn)
{
	const unsigned char EMPTY_MAC[6] = { 0x00,0x00,0x00,0x00,0x00,0x00 };
	FILE *fp;
	int i = 0;
	int p = 0;
	char buffer[1024];
	
	
	if ((fp = fopen(BLT_DEVICE_NAMES_FILE, "w")) == NULL) {
		return(0);
	}

	while (i < dn->numDevices && p < dn->maxNumDevices) {
		if (memcmp(dn->devices[p].bdAddr, EMPTY_MAC, 6) != 0) {
			
			memset(buffer, 0, 1024);
			printWithEsc(dn->devices[p].name, buffer, 1024);
			
			fprintf(fp, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X \"%s\"\n",
				dn->devices[p].bdAddr[0],dn->devices[p].bdAddr[1],dn->devices[p].bdAddr[2],
				dn->devices[p].bdAddr[3],dn->devices[p].bdAddr[4],dn->devices[p].bdAddr[5],
				buffer);
			i++;
		}
		p++;
	}
	
	fclose(fp);
	return(1);
}
#endif

int blt_dn_load(struct blt_dev_names_t *dn)
{
	FILE *fp;
	struct blt_device_t dev;
	unsigned int a, b, c, d, e, f;
	char buffer[1024];
	char *p;
	
	
	if ((fp = fopen(BLT_DEVICE_NAMES_FILE, "r")) == NULL) {
		return(0);
	}

	while (!feof(fp)) {
		memset(&dev, 0, sizeof(struct blt_device_t));
		if (fgets(buffer, 1024, fp) == NULL) break;
		
		sscanf(buffer, "%X:%X:%X:%X:%X:%X",&a,&b,&c,&d,&e,&f);
		dev.bdAddr[0] = (unsigned char) a;
		dev.bdAddr[1] = (unsigned char) b;
		dev.bdAddr[2] = (unsigned char) c;
		dev.bdAddr[3] = (unsigned char) d;
		dev.bdAddr[4] = (unsigned char) e;
		dev.bdAddr[5] = (unsigned char) f;
		
		p = strstr(buffer, "\"");
		if (p > 0) scanWithEsc(p+1, dev.name, BLT_DEVICE_NAME_LENGTH+1);
		else continue;
		
		blt_dn_write(dn, &dev);
	}
	fclose(fp);
	return(1);
}
