/*-------------------------------------------------------------------
 *
 *  @file push.c
 *
 *  @author Collin R. Mulliner <collin@mulliner.org>
 *
 *  (c) Collin R. Mulliner
 *
 *  This is part of the Bluetooth Location Tracker (BLT) project.
 *
 *-------------------------------------------------------------------*/

/*
 * This file is part of blt_server
 *
 * blt_server is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * blt_server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blt_server; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include "blt_server.h"

/**
 *  @brief init push server structure
 *
 *  @param p pointer to struct
 *
 *  @returns nothing
 */
void blt_push_init(struct blt_pusher_t *p)
{
	assert(p != NULL);
	
	p->first = NULL;
	p->last = NULL;
	
	pthread_mutex_init(&p->lock, NULL);
	pthread_cond_init(&p->sig, NULL);
	
	p->status = 1;
}

/**
 *  @brief destroy push server structure
 *
 *  @param p pointer to structure
 *
 *  @returns nothing
 */
void blt_push_destroy(struct blt_pusher_t *p)
{
	struct blt_push_t *e;
	
	
	assert(p != NULL);
	p->status = 0;

	/* free memory */
	while (blt_push_get(p, &e)) {
		assert(e != NULL);
		free(e);
	}
	
	pthread_mutex_destroy(&p->lock);
	pthread_cond_destroy(&p->sig);
}

/**
 *  @brief put push event to (start of) queue
 *
 *  @param p pointer to push server
 *  @param n pointer to event
 *
 *  @returns nothing
 */
void blt_push_put(struct blt_pusher_t *p, struct blt_push_t *n)
{
	assert(p != NULL);
	assert(n != NULL);

	pthread_mutex_lock(&p->lock);
	
	if (p->first == NULL) {
		p->first = n;
		p->last = n;
		n->prev = NULL;
	}
	else if (p->first != NULL) {
		n->prev = NULL;
		p->first->prev = n;
		p->first = n;
	}
	else {
		#ifdef DEBUG
		fprintf(stderr, "push Server: broken!\n");
		#endif
		p->first = NULL;
		p->last = NULL;
	}

	pthread_mutex_unlock(&p->lock);
}

/**
 *  @brief get last (oldest) event from queue
 *
 *  @param p pointer to push server
 *  @param n pointer to pointer to push event
 *
 *  @returns 1 on success or 0 if no event is available
 */
int blt_push_get(struct blt_pusher_t *p, struct blt_push_t **n)
{
	int retval;


	assert(p != NULL);
	assert(n != NULL);
	
	if (p->last != NULL) {
		*n = p->last;
		p->last = (*n)->prev;
		if (p->last == NULL) p->first = NULL;
		retval = 1;
	}
	else {
		retval = 0;
	}
	
	return(retval);
}

/**
 *  @brief the push server thread
 *
 *  The tracker thread push their events to the push server and
 *  the push server distributes the event to all push clients.
 *
 *  @param pointer to server session
 *
 *  @returns nothing
 */
void *blt_pusher(void *args)
{
	struct blt_server_session_t *server; /* = (struct blt_server_session_t*) args; */
	struct blt_pusher_t *self; /* = &server->pushServer; */
	struct blt_push_t *event = NULL;
	struct blt_push_t *newEvent = NULL;
	int i, i2;
	
	
	assert(args != NULL);
	server = (struct blt_server_session_t*) args;
	self = &server->pushServer;
	assert(self != NULL);
	
	do {
		if (event == NULL) {
			#ifdef DEBUG3
			fprintf(stderr, "pushServer: sleeping\n");
			#endif
			pthread_cond_wait(&self->sig, &self->lock);
			#ifdef DEBUG3
			fprintf(stderr, "pushServer: woke up\n");
			#endif
			blt_push_get(self, &event);
			pthread_mutex_unlock(&self->lock);
			continue;
		}
		else {
			#ifdef DEBUG3
			fprintf(stderr, "pushServer: distributing event to push clients\n");
			#endif
		
			/* notify all pushing data clients */
			i = 0; i2 = 0;
			do {
				if (server->datas[i].myListIndex != -1) {
					/* only add event if client is pushing */
					if (server->datas[i].isPush) {
						newEvent = malloc(sizeof(struct blt_push_t));
						memcpy(newEvent, event, sizeof(struct blt_push_t));
						blt_push_put(&server->datas[i].myPush, newEvent);
						newEvent = NULL;
						/* signal thread */
						pthread_cond_signal(&server->datas[i].myPush.sig);
					}
					i2++;
				}
				i++;
			} while(i < MAX_NUM_DATAS && i2 < server->numDatas);
			
			#ifdef DEBUG3
			fprintf(stderr, "pushServer: event distribution completed\n");
			#endif
			
			if (event != NULL) {
				free(event);
				event = NULL;
			}
			
			pthread_mutex_lock(&self->lock);
			blt_push_get(self, &event);
			pthread_mutex_unlock(&self->lock);
		}
	} while (self->status);
		
	return(NULL);
}
